@@ -8,17 +8,19 @@
 #' @param stdout Standard output, FALSE to ignore, TRUE for temp file.
 #' @param stderr Standard error, FALSE to ignore, TRUE for temp file.
 #' @param cleanup Kill on GC?
+#' @param wd working directory (or NULL)
 #' @param echo_cmd Echo command before starting it?
 #' @param supervise Should the process be supervised?
 #' @param encoding Assumed stdout and stderr encoding.
+#' @param post_process Post processing function.
 #'
 #' @keywords internal
 #' @importFrom utils head tail
 
 process_initialize <- function(self, private, command, args,
-                               stdout, stderr, cleanup,
+                               stdout, stderr, cleanup, wd,
                                echo_cmd, supervise, windows_verbatim_args,
-                               windows_hide_window, encoding) {
+                               windows_hide_window, encoding, post_process) {
 
   "!DEBUG process_initialize `command`"
 
@@ -27,20 +29,24 @@ process_initialize <- function(self, private, command, args,
   assert_that(is_string_or_null(stdout))
   assert_that(is_string_or_null(stderr))
   assert_that(is_flag(cleanup))
+  assert_that(is_string_or_null(wd))
   assert_that(is_flag(echo_cmd))
   assert_that(is_flag(windows_verbatim_args))
   assert_that(is_flag(windows_hide_window))
   assert_that(is_string(encoding))
+  assert_that(is.function(post_process) || is.null(post_process))
 
   private$command <- command
   private$args <- args
   private$cleanup <- cleanup
+  private$wd <- wd
   private$pstdout <- stdout
   private$pstderr <- stderr
   private$echo_cmd <- echo_cmd
   private$windows_verbatim_args <- windows_verbatim_args
   private$windows_hide_window <- windows_hide_window
   private$encoding <- encoding
+  private$post_process <- post_process
 
   if (echo_cmd) do_echo_cmd(command, args)
 
@@ -49,7 +55,7 @@ process_initialize <- function(self, private, command, args,
     c_processx_exec,
     command, c(command, args), stdout, stderr,
     windows_verbatim_args, windows_hide_window,
-    private, cleanup, encoding
+    private, cleanup, wd, encoding
   )
   private$starttime <- Sys.time()
 
