@@ -14,10 +14,10 @@ NULL
 #' ```
 #' p <- process$new(command = NULL, args,
 #'                  stdout = NULL, stderr = NULL, cleanup = TRUE,
-#'                  echo_cmd = FALSE, supervise = FALSE,
+#'                  wd = NULL, echo_cmd = FALSE, supervise = FALSE,
 #'                  windows_verbatim_args = FALSE,
 #'                  windows_hide_window = FALSE,
-#'                  encoding = "")
+#'                  encoding = "", post_process = NULL)
 #'
 #' p$is_alive()
 #' p$signal(signal)
@@ -43,6 +43,8 @@ NULL
 #'
 #' p$poll_io(timeout)
 #'
+#' p$get_result()
+#'
 #' print(p)
 #' ```
 #'
@@ -63,6 +65,8 @@ NULL
 #'     `"|"`: create a connection for it.
 #' * `cleanup`: Whether to kill the process (and its children)
 #'     if the `process` object is garbage collected.
+#' * `wd`: working directory of the process. It must exist. If `NULL`, then
+#'     the current working directory is used.
 #' * `echo_cmd`: Whether to print the command to the screen before
 #'     running it.
 #' * `supervise`: Whether to register the process with a supervisor.
@@ -84,6 +88,9 @@ NULL
 #'     both streams in UTF-8 currently. If you want to read them
 #'     without any conversion, on all platforms, specify `"UTF-8"` as
 #'     encoding.
+#' * `post_process`: An optional function to run when the process has
+#'     finished. Currently it only runs if `$get_result()` is called.
+#'     It is only run once.
 #'
 #' @section Details:
 #' `$new()` starts a new process in the background, and then returns
@@ -108,12 +115,21 @@ NULL
 #' `$wait()` waits until the process finishes, or a timeout happens.
 #' Note that if the process never finishes, and the timeout is infinite
 #' (the default), then R will never regain control. It returns
-#' the process itself, invisibly.
+#' the process itself, invisibly. In some rare cases, `$wait()` might take
+#' a bit longer than specified to time out. This happens on Unix, when
+#' another package overwrites the processx SIGCHLD signal handler, after the
+#' processx process has started. One such package is parallel, if used
+#' with fork clusters, e.g. through `parallel::mcparallel()`.
 #'
 #' `$get_pid()` returns the process id of the process.
 #'
 #' `$get_exit_status` returns the exit code of the process if it has
-#' finished and `NULL` otherwise.
+#' finished and `NULL` otherwise. On Unix, in some rare cases, the exit
+#' status might be `NA`. This happens if another package (or R itself)
+#' overwrites the processx SIGCHLD handler, after the processx process
+#' has started. In these cases processx cannot determine the real exit
+#' status of the process. One such package is parallel, if used with
+#' fork clusters, e.g. through the `parallel::mcparallel()` function.
 #'
 #' `$restart()` restarts a process. It returns the process itself.
 #'
@@ -209,6 +225,11 @@ NULL
 #' the _Polling_ section, and see also the [poll()] function
 #' to poll on multiple processes.
 #'
+#' `$get_result()` returns the result of the post processesing function.
+#' It can only be called once the process has finished. If the process has
+#' no post-processing function, then `NULL` is returned.
+#'
+#'
 #' `print(p)` or `p$print()` shows some information about the
 #' process on the screen, whether it is running and it's process id, etc.
 #'
@@ -249,13 +270,13 @@ process <- R6Class(
   public = list(
 
     initialize = function(command = NULL, args = character(),
-      stdout = NULL, stderr = NULL, cleanup = TRUE,
+      stdout = NULL, stderr = NULL, cleanup = TRUE, wd = NULL,
       echo_cmd = FALSE, supervise = FALSE, windows_verbatim_args = FALSE,
-      windows_hide_window = FALSE, encoding = "")
+      windows_hide_window = FALSE, encoding = "",  post_process = NULL)
       process_initialize(self, private, command, args,
-                         stdout, stderr, cleanup, echo_cmd, supervise,
+                         stdout, stderr, cleanup, wd, echo_cmd, supervise,
                          windows_verbatim_args, windows_hide_window,
-                         encoding),
+                         encoding, post_process),
 
     kill = function(grace = 0.1)
       process_kill(self, private, grace),
@@ -341,7 +362,10 @@ process <- R6Class(
       process_get_error_file(self, private),
 
     poll_io = function(timeout)
-      process_poll_io(self, private, timeout)
+      process_poll_io(self, private, timeout),
+
+    get_result = function()
+      process_get_result(self, private)
   ),
 
   private = list(
@@ -354,6 +378,7 @@ process <- R6Class(
     pstdout = NULL,       # the original stdout argument
     pstderr = NULL,       # the original stderr argument
     cleanfiles = NULL,    # which temp stdout/stderr file(s) to clean up
+    wd = NULL,            # working directory (or NULL for current)
     starttime = NULL,     # timestamp of start
     echo_cmd = NULL,      # whether to echo the command
     windows_verbatim_args = NULL,
@@ -371,6 +396,10 @@ process <- R6Class(
 
     encoding = "",
 
+    post_process = NULL,
+    post_process_result = NULL,
+    post_process_done = FALSE,
+
     get_short_name = function()
       process_get_short_name(self, private)
   )
@@ -410,11 +439,13 @@ process_restart <- function(self, private) {
     private$pstdout,
     private$pstderr,
     private$cleanup,
+    private$wd,
     private$echo_cmd,
     private$supervised,
     private$windows_verbatim_args,
     private$windows_hide_window,
-    private$encoding
+    private$encoding,
+    private$post_process
   )
 
   invisible(self)
@@ -495,3 +526,12 @@ process_supervise <- function(self, private, status) {
     private$supervised <- FALSE
   }
 }
+
+process_get_result <- function(self, private) {
+  if (self$is_alive()) stop("Process is still alive")
+  if (!private$post_process_done && is.function(private$post_process)) {
+    private$post_process_result <- private$post_process()
+    private$post_process_done <- TRUE
+  }
+  private$post_process_result
+}
